package org.n52.wps.client.jump;

import java.awt.Component;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTabbedPane;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.JTree;
import javax.swing.JViewport;
import javax.swing.ToolTipManager;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.TreeSelectionModel;

import net.opengis.wps.x100.CapabilitiesDocument;
import net.opengis.wps.x100.ExecuteDocument;
import net.opengis.wps.x100.ExecuteResponseDocument;
import net.opengis.wps.x100.InputDescriptionType;
import net.opengis.wps.x100.OutputDescriptionType;
import net.opengis.wps.x100.ProcessBriefType;
import net.opengis.wps.x100.ProcessDescriptionType;

import org.apache.log4j.Logger;
import org.n52.wps.client.ExecuteRequestBuilder;
import org.n52.wps.client.ExecuteResponseAnalyser;
import org.n52.wps.client.WPSClientException;
import org.n52.wps.client.WPSClientSession;

import com.vividsolutions.jump.feature.FeatureCollection;
import com.vividsolutions.jump.workbench.WorkbenchContext;
import com.vividsolutions.jump.workbench.model.Category;
import com.vividsolutions.jump.workbench.model.Layer;
import com.vividsolutions.jump.workbench.model.Layerable;
import com.vividsolutions.jump.workbench.model.StandardCategoryNames;
import com.vividsolutions.jump.workbench.plugin.AbstractPlugIn;
import com.vividsolutions.jump.workbench.plugin.PlugInContext;
import com.vividsolutions.jump.workbench.ui.MultiInputDialog;

public class WPSClientExecuteProcessPlugIn extends AbstractPlugIn {

	private static Logger LOGGER = Logger.getLogger(WPSClientExecuteProcessPlugIn.class);
	private static final String DESCRIPTION_TAB = "DescribeTab";
	private static final String EXECUTE_PROCESS_TAB = "ExecuteProcessTab";
	WorkbenchContext workbenchContext;
	private MultiInputDialog dialog;
	private WPSClientSession client;
	
	//FeatureType type;
	
	public boolean execute(PlugInContext ctx) throws Exception {
		dialog = new MultiInputDialog(ctx.getWorkbenchFrame(), "Execute Process Wizard", true);
		dialog.setSize(800, 600);
		workbenchContext = ctx.getWorkbenchContext();
		client = (WPSClientSession)ctx.getWorkbenchContext().getBlackboard().get(WPSClientExtension.WPS_Client);
		JSplitPane splitPane = new JSplitPane();
		JTabbedPane tabbedPane = new JTabbedPane();
		JTree tree = null;
		// Defining the description Panel
		JPanel describePanel = new JPanel(new GridLayout(1,1));
		JTextArea textArea = new JTextArea();
		textArea.setEditable(false);
		JScrollPane scrollPaneDescription = new JScrollPane(textArea, JScrollPane.VERTICAL_SCROLLBAR_ALWAYS, JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);
		describePanel.add(scrollPaneDescription);
		describePanel.setName(DESCRIPTION_TAB);
		// defining the execute Panel
		JPanel executePanel = new JPanel(new GridBagLayout());
		GridBagConstraints c = new GridBagConstraints();
		JPanel executeParameterPanel = new JPanel(new GridLayout(1,1));
		JScrollPane scrollExecutePane = new JScrollPane(executeParameterPanel, JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED, JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
		c.anchor = GridBagConstraints.NORTH;
		c.fill = GridBagConstraints.NONE;
		c.gridwidth = 2;
		c.gridx = 0;
		c.gridy = 0;
		c.ipady = 50;
		executePanel.add(scrollExecutePane, c);
		c.anchor = GridBagConstraints.PAGE_END;
		JButton executeProcessButton = new JButton("Execute Process");
		c.gridx = 1;
		c.gridy = 1;
		c.ipady = 0;
		c.gridwidth = 1;
		executePanel.add(executeProcessButton, c);
		executePanel.setName(EXECUTE_PROCESS_TAB);
		tabbedPane.addTab("Describe", null, describePanel, "was sonst");
		tabbedPane.addTab("Execute", null, executePanel, "und sonst");
		tree = initTree(tabbedPane);
		tree.addTreeSelectionListener(new MyTreeSelectionListener(tree, tabbedPane));
		tabbedPane.addMouseListener(new MyTabMouseListener(tree, tabbedPane));
		JScrollPane scrollPaneTree = new JScrollPane(tree);
		executeProcessButton.addMouseListener(new ExecuteProcessListener(executePanel, tree));
		splitPane.setOrientation(JSplitPane.HORIZONTAL_SPLIT);
		splitPane.setDividerLocation(150);
		splitPane.setOneTouchExpandable(true);
		splitPane.setLeftComponent(scrollPaneTree);
		splitPane.setRightComponent(tabbedPane);
		dialog.add(splitPane);
		dialog.setSize(800, 600);
		dialog.setResizable(true);
		dialog.setVisible(true);
		if(dialog.wasOKPressed()){
			return false;
		}
		return true;
	}

	public void initialize(PlugInContext ctx) throws Exception {
		ctx.getFeatureInstaller().addMainMenuItem(this, WPSClientExtension.WPS_CLIENT_MENU_PATH, "Execute Process", null, null);
		
	}
	
	private class ExecuteProcessListener implements MouseListener {
		
		private JPanel executePanel;
		private JTree tree;

		public ExecuteProcessListener(JPanel executePanel, JTree tree) {
			this.executePanel = executePanel;
			this.tree = tree;
		}
		public void mouseClicked(MouseEvent arg0) {
			try {
				executeProcess(executePanel, tree);
			}
			catch(IOException e) {
				e.printStackTrace();
			}
		}
		
		public void mousePressed(MouseEvent arg0) {
		}

		public void mouseReleased(MouseEvent arg0) {
		}

		public void mouseEntered(MouseEvent arg0) {
		}

		public void mouseExited(MouseEvent arg0) {
		}
	}
	
	private void executeProcess(JPanel executePanel, JTree tree) throws IOException {
		Component[] components = ((JPanel)((JViewport)((JScrollPane)executePanel.getComponent(0)).getComponent(0)).getComponent(0)).getComponents();
		ExecuteDocument doc = ExecuteDocument.Factory.newInstance();
		DefaultMutableTreeNode selectedNode = (DefaultMutableTreeNode) tree.getLastSelectedPathComponent();
		ProcessBean process = (ProcessBean) selectedNode.getUserObject();
		String serverID = process.getServerID();
		String processID = process.getProcess().getIdentifier().getStringValue();
		ProcessDescriptionType processDesc = client.getProcessDescription(serverID, processID);
		ExecuteRequestBuilder erb = new ExecuteRequestBuilder(processDesc);
		for(Component component : components) {
			if(! (component instanceof JLabel)) {
				String parameterName = component.getName();
				if(component instanceof JComboBox) {
					Layer layer = (Layer)((JComboBox)component).getSelectedItem();
					FeatureCollection fc = layer.getFeatureCollectionWrapper();
					org.geotools.feature.FeatureCollection gtCollection = null;
					try {
						 gtCollection = JUMPConversionUtil.convert2GTFeatures(fc);
					}
					catch(Exception e) {
						LOGGER.warn(e);
					}
					erb.addComplexData(parameterName, gtCollection);

				}
				else if(component instanceof JTextField) {
					String parameter = ((JTextField)component).getText();
					erb.addLiteralData(parameterName, parameter);
				}
			}
		}
		try {
			ExecuteResponseDocument executeResponse = null;
			Object response = client.execute(serverID, erb.getExecute());
			if (response instanceof ExecuteResponseDocument) {
				executeResponse = (ExecuteResponseDocument) response;
			}
			if(executeResponse != null) {
				handleResponse(executeResponse, processDesc);
				workbenchContext.getWorkbench().getFrame().log("ProcessExecute successful!");
			}
			else {
				LOGGER.warn("executeResponse is null for process: " + processID);
			}
		}
		catch(WPSClientException e) {
			workbenchContext.getWorkbench().getFrame().log(e.getMessage());
			if(e.isServerException()) {
				String serverError = e.getServerException().toString();
				workbenchContext.getWorkbench().getFrame().log("Error send by the server: " + serverError);
			}
			else if(e.getCause() != null) {
				String error = e.getCause().getMessage();
				workbenchContext.getWorkbench().getFrame().log(error);
			}
			LOGGER.warn(e);
			workbenchContext.getWorkbench().getFrame().warnUser("Error occured while processing, please check Log");
		}
	}
	
	// Supports only complexValue response
	private void handleResponse(ExecuteResponseDocument executeResponse, ProcessDescriptionType processDesc) throws WPSClientException{
		ExecuteResponseAnalyser era = new ExecuteResponseAnalyser(executeResponse, processDesc);
		OutputDescriptionType[] outputs = processDesc.getProcessOutputs().getOutputArray();
		for(OutputDescriptionType output : outputs) {
			if(output.isSetComplexOutput()) {
				org.geotools.feature.FeatureCollection gtCollection = (org.geotools.feature.FeatureCollection)era.getComplexData(output.getIdentifier().getStringValue());
				FeatureCollection fc = null;
				try {
					fc = JUMPConversionUtil.convert2JUMPFeatures(gtCollection);
				}
				catch(RuntimeException e) {
					workbenchContext.getWorkbench().getFrame().warnUser("Error while parsing result features");
					throw new WPSClientException("error while parsing", e);
				}
				if(fc != null) {
					workbenchContext.getLayerManager().addLayer(StandardCategoryNames.WORKING, "Process Result", fc);
				}
			}
		}
	}

	private InputDescriptionType getInputDescription(String identifier, ProcessDescriptionType dp) {
		InputDescriptionType[] inputs = dp.getDataInputs().getInputArray();
		for(InputDescriptionType input : inputs) {
			if(input.getIdentifier().getStringValue().equals(identifier)) {
				return input;
			}
		}
		return null;
	}
	
	private JTree initTree(JTabbedPane pane) {
		DefaultMutableTreeNode root = new DefaultMutableTreeNode("Logged WPS");
		for(String serverID : client.getLoggedServices()) {
			DefaultMutableTreeNode serverNode = new DefaultMutableTreeNode(serverID);
			CapabilitiesDocument caps = client.getWPSCaps(serverID);
			ProcessBriefType[] processes = caps.getCapabilities().getProcessOfferings().getProcessArray();
			for(ProcessBriefType process : processes) {
				DefaultMutableTreeNode processNode = new DefaultMutableTreeNode(process.getIdentifier().getStringValue());
				processNode.setUserObject(new ProcessBean(process, serverID));
				serverNode.add(processNode);
			}
			root.add(serverNode);
		}
		JTree tree = new JTree(root);
		ToolTipManager.sharedInstance().registerComponent(tree);
		tree.getSelectionModel().setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);		
		return tree;
	}
	
	private class MyTreeSelectionListener implements TreeSelectionListener{
		JTree t;
		JTabbedPane tabPane;
		MyTreeSelectionListener(JTree t, JTabbedPane tabPane) {
			this.t = t;
			this.tabPane = tabPane;
		}
		public void valueChanged(TreeSelectionEvent e) {
			updateTab(t, tabPane);
			tabPane.updateUI();
		}
	}
	
	private class MyTabMouseListener implements MouseListener {

		private JTabbedPane tabPane;
		private JTree tree;

		MyTabMouseListener(JTree tree, JTabbedPane tabPane) {
			this.tree = tree;
			this.tabPane = tabPane;
		}

		public void mouseClicked(MouseEvent arg0) {
			updateTab(tree, tabPane);
			tabPane.updateUI();
		}

		public void mousePressed(MouseEvent arg0) {
			
		}

		public void mouseReleased(MouseEvent arg0) {
			
		}

		public void mouseEntered(MouseEvent arg0) {
			
		}

		public void mouseExited(MouseEvent arg0) {
			
		}
	}
	
	private void updateTab(JTree tree, JTabbedPane tabPane) {
		DefaultMutableTreeNode node = (DefaultMutableTreeNode) tree.getLastSelectedPathComponent();
		if(node == null) {
			return;
		}
		if(!node.isRoot() && node.getUserObject() instanceof ProcessBean) {
			ProcessBean process = (ProcessBean)node.getUserObject();
			try {
				setTabbedPane(process, tabPane);
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		else if(!node.isRoot() && !(node.getUserObject() instanceof ProcessBean)) {
			setTabbedPane((String)node.getUserObject(), tabPane);
		}
	}
	
	/**
	 * sets the tabbedPane if a serverNode is activiated
	 * @param string
	 * @param tabPane
	 */
	private void setTabbedPane(String serverID, JTabbedPane tabPane) {
		JPanel selectedTab = (JPanel) tabPane.getSelectedComponent();
		CapabilitiesDocument caps = client.getWPSCaps(serverID);
		if(selectedTab.getName().equals(DESCRIPTION_TAB)) {
			JTextArea textArea = (JTextArea)((javax.swing.JViewport)((JScrollPane)selectedTab.getComponent(0)).getComponent(0)).getComponent(0);
			String capabilitiesText = caps.toString();
			textArea.setText(capabilitiesText);
		}
		else if(selectedTab.getName().equals(EXECUTE_PROCESS_TAB)) {
			setVisiblity(selectedTab, false);
		}
	}

	private void setVisiblity(JPanel pane, boolean visibility) {
		for(Component component : pane.getComponents()) {
			component.setVisible(visibility);
		}
	}
	
	private class ProcessBean{
		ProcessBriefType p;
		String serverID;
		
		ProcessBean(ProcessBriefType p, String serverID) {
			this.p = p;
			this.serverID = serverID;
		}
		
		public ProcessBriefType getProcess() {
			return p;
		}

		public String getServerID() {
			return serverID;
		}

		public String toString() {
			String processTitle = p.getTitle().getStringValue();
			if(processTitle == null || processTitle.equals("")) {
				return p.getIdentifier().getStringValue();
			}
			return processTitle;
		}
	}
	
	private void setTabbedPane(ProcessBean process, JTabbedPane tabPane) throws IOException {
		JPanel selectedTab = (JPanel) tabPane.getSelectedComponent();
		String processID = process.getProcess().getIdentifier().getStringValue();
		String serverID = process.getServerID();
		ProcessDescriptionType processDesc = client.getProcessDescription(serverID, processID);
		if(selectedTab.getName().equals(DESCRIPTION_TAB)) {
			JTextArea textArea = (JTextArea)((javax.swing.JViewport)((JScrollPane)selectedTab.getComponent(0)).getComponent(0)).getComponent(0);
			String processDescText = processDesc.toString();
			textArea.setText(processDescText);
		}
		else if(selectedTab.getName().equals(EXECUTE_PROCESS_TAB)) {
			setVisiblity(selectedTab, true);
			JViewport viewPort = (JViewport)((JScrollPane)selectedTab.getComponent(0)).getComponent(0);
			JPanel parameterPanel = setupParameterPanel(processDesc);
			viewPort.add(parameterPanel);
		}
	}

	private JPanel setupParameterPanel(ProcessDescriptionType processDesc) {
		JPanel panel = new JPanel();
		panel.setLayout(new GridBagLayout());
		InputDescriptionType[] inputs = processDesc.getDataInputs().getInputArray();
		GridBagConstraints c = new GridBagConstraints();
		c.anchor = GridBagConstraints.WEST;
		int rowPosition = 0;
		for(InputDescriptionType input : inputs) {
			String inputName = input.getIdentifier().getStringValue();
			 JLabel label = new JLabel(inputName);
			 JComponent component = null;
			 if(input.isSetComplexData()) {
				 component = new JComboBox();
				 List<Category> categories = workbenchContext.getLayerManager().getCategories();
				 List<String> layerNames = new ArrayList<String>();
				 for(Category cat : categories) {
					 List<Layerable> layerables = cat.getLayerables();
					 for(Layerable layerable : layerables) {
						 if(layerable instanceof Layer) {
							 layerNames.add(layerable.getName());
							 ((JComboBox)component).addItem(layerable);
						 }
					 }
				 }
				 if(layerNames.size() != 0) {
					 ((JComboBox) component).setSelectedIndex(0);
				 }
			 }
			 else if(input.isSetLiteralData()) {
				 component = new JTextField(20);
			 }
			 if(component != null) {
				component.setName(inputName);
				c.fill = GridBagConstraints.NONE;
				c.gridx = 1;
				c.anchor = GridBagConstraints.FIRST_LINE_START;
				c.gridy = rowPosition;
				panel.add(component, c);
			}
//			label.setHorizontalAlignment(JLabel.LEFT);
			c.gridx = 0;
			c.gridy = rowPosition;
			c.anchor = GridBagConstraints.WEST;
			panel.add(label, c);
			rowPosition++;
			}
		return panel;
	}
}


